
//Script version 3.7.7
(function scriptedWebServiceOperation(request, response) {
	
	var skipCount=0;
	var updateCount=0;
	var createCount=0;
	var deleteCount=0;
        var maxAttachSize=request.maxAttachSize;
	var startTimestamp=Math.floor(Date.now() / 1000); //timestamp in seconds
        var useGlide=typeof(request.sourceURL)==="undefined" || request.sourceURL===null || request.sourceURL.equals('');
	try {
            response.log_entries="";
            response.processed_target_list="";
            response.oversize_attachment_warnings="";
            response.oversize_attachment_files="";
            addToLog(response,request.target_table +": attachment migration ["+request.target_table+"]");
            if(useGlide)
                addToLog(response,"Use Glide API to retrieve attachments");
            else
                addToLog(response,"Use REST Attachment API to retrieve attachments ["+ request.sourceURL +"]");
			addToLog(response,"Maximum Attachment Size: " + maxAttachSize);
            var glideSysAttachment =new GlideSysAttachment();
            var targetIDSet=request.target_id_list.split(",");
            var sourceIDSet=request.source_id_list.split(",");
            if(targetIDSet.length!==sourceIDSet.length) {
                response.error_msg="Number of target entries ["+ targetIDSet.length +"] does not match number of source entries [" + sourceIDSet.length+"]";
                response.status="ERROR";
                return;
            }
			//build the image field mappings
            var imageFieldMap=buildImageFieldMap(request.source_image_list,request.target_image_list);
			//retrieve a map of all image fields source and target
            var targetImageFields=getImageFieldListGlide(response,request.target_table);
			var sourceImageFields;
            if(useGlide) {
                sourceImageFields=getImageFieldListGlide(response,request.source_table);
            }
            else {
                sourceImageFields=getImageFieldListREST(response,request.sourceURL,request.authorization,request.source_table);
            }
			//loop through entries and process the attachment migration for each
            for(var i=0;i<targetIDSet.length;i++) {
                if(Math.floor(Date.now() / 1000)>startTimestamp+30) {
                    response.status="TIMEOUT";
                    response.error_msg="Timeout before completion (>30 seconds elapsed)";
                    return;
                }
                addToLog(response,"Processing attachments for source entry ["+ sourceIDSet[i]+"]");
                //get the targetRecord
                var grTarget=new GlideRecord(request.target_table);
                if(!grTarget.get("sys_id", targetIDSet[i])) {
                    response.error_msg="Target Entry " + targetIDSet[i]+ " not found";
                    response.status="ERROR";
                    return;
                }
                // get the source and target attachments
                var tgtAttachments=getAttachmentMetaData_Glide(response,request.target_table,targetIDSet[i],targetImageFields);
                var srcAttachments;
                if(useGlide) {
                    srcAttachments=getAttachmentMetaData_Glide(response,request.source_table,sourceIDSet[i],sourceImageFields);
                }
                else {
                    srcAttachments=getAttachmentMetaData_REST(response,request.sourceURL,request.source_table,sourceIDSet[i],sourceImageFields,request.authorization);
                }
                //process each source attachment found
                for(var key in srcAttachments) {
                    var srcAttachRecord=srcAttachments[key];
                    var action;
                    var type="Attachment";
                    var tgtAttachName=srcAttachRecord.file_name;
                    if(srcAttachRecord.isImage) {
                        type="Image";
                        tgtAttachName=imageFieldMap[srcAttachRecord.file_name];
                    }
					var tgtAttachRecord=getAttachmentWithName(tgtAttachments,tgtAttachName);
                   if(tgtAttachName===undefined) {
                        action="IGNORE"; //unmapped image
                    }
                    else if(!srcAttachRecord.isImage && !request.migrate_attachments.equals("TRUE"))
                        action="IGNORE"; //attachments are not to be migrated
                    else {
                        if(tgtAttachRecord!==undefined ){
                            if(tgtAttachRecord["size_bytes"].equals(srcAttachRecord.size_bytes)) {
                                action="SKIP";
                            }
                            else {
                                action="UPDATE";
                            }
                        }
                        else {
                            action="CREATE";
                        }
                        //overwrite the action if the file is too big
                        if(action!=="SKIP" && Number(srcAttachRecord.size_bytes)>Number(maxAttachSize)) {
                            action="OVERSIZE";
                        }
                    }
                    switch(action) {
                        case "SKIP":
                            skipCount++;
                            addToLog(response,"Skip Duplicate "+type+": " + srcAttachRecord.file_name);
                            break;
                        case "UPDATE":
                            //add new attachment
                            var updatedAttachID;
                            if(useGlide) {
                                updatedAttachID=setAttachment_Glide(response,glideSysAttachment, grTarget,srcAttachRecord.sys_id,srcAttachRecord.content_type,tgtAttachName,srcAttachRecord.isImage);
                            }
                            else {
                                updatedAttachID=setAttachment_REST(response, request.sourceURL,srcAttachRecord.sys_id,request.authorization,request.target_table,grTarget.sys_id,tgtAttachName,srcAttachRecord.isImage);
                            }
                            addToLog(response,"Added Updated "+type+":" + tgtAttachName+' [' +updatedAttachID + ']');
                            //delete original
                            glideSysAttachment.deleteAttachment(tgtAttachRecord["sys_id"]);
                            addToLog(response,"Deleted Original "+type+":" + tgtAttachName);
                            updateCount++;
                            break;
                        case "CREATE":
                            var newAttachID;
                        if(useGlide) {
                            newAttachID=setAttachment_Glide(response,glideSysAttachment, grTarget,srcAttachRecord.sys_id,srcAttachRecord.content_type,tgtAttachName,srcAttachRecord.isImage);
                        }
                        else {
                            newAttachID=setAttachment_REST(response, request.sourceURL,srcAttachRecord.sys_id,request.authorization,request.target_table,grTarget.sys_id,tgtAttachName,srcAttachRecord.isImage);
                        }					
                        addToLog(response,"Create "+type+": " + srcAttachRecord.file_name +' [' +newAttachID + ']');
                        createCount++;
                        break;
                        case "OVERSIZE":
                            skipCount++;
							addToLog(response,"WARNING: Oversize "+type+" skipped [" + srcAttachRecord.file_name+"]");
                            if(!response.oversize_attachment_warnings.equals(""))response.oversize_attachment_warnings+=",";
                                response.oversize_attachment_warnings+=targetIDSet[i];
							if(!response.oversize_attachment_files.equals(""))response.oversize_attachment_files+=",";
                                response.oversize_attachment_files+=srcAttachRecord.file_name;
							break;
                    }
                }
				//finally check for NULL target attachments that are mapped
                for(var key in imageFieldMap) {
                    var tgtAttachRecord=getAttachmentWithName(tgtAttachments,imageFieldMap[key]);
                    if(getAttachmentWithName(srcAttachments,key)===undefined && tgtAttachRecord!==undefined) {
                        //null image migration, delete target
                        addToLog(response,"delete: "+ tgtAttachRecord["file_name"]);
						deleteCount++;
                        glideSysAttachment.deleteAttachment(tgtAttachRecord["sys_id"]);
                    }
                }
				//attachment migration complete, add sys id to processed 
				addToLog(response,"Attachment/Image Migration Complete.");
				if(!response.processed_target_list.equals(""))response.processed_target_list+=",";
                        response.processed_target_list+=targetIDSet[i];
            }
            response.status="COMPLETE";
	}
	catch(e) {
            response.status="ERROR";
            if(e!=null) {
                addToLog(response,e);
                response.error_msg=e;
            }
	}
	response.update_count=updateCount.toFixed();
	response.create_count=createCount.toFixed();
	response.skip_count=skipCount.toFixed();
	addToLog(response,"Created :" + createCount.toFixed()+" Updated :" +updateCount.toFixed()+" Deleted :"+deleteCount.toFixed()+" Skipped :"+ skipCount.toFixed());
	
	
})(request, response);

function addToLog(response,msg) {
	response.log_entries+=msg+"\n";
}

//returns a list of source attachments and images associated with this record
function getAttachmentMetaData_REST(response, instanceURL, sourceTable, sourceSysId,imageFieldsMap,  authorization) {
    var getMetaDataRequest = new sn_ws.RESTMessageV2();
    var attachmentMetaData={};
    getMetaDataRequest.setHttpMethod('GET');
    getMetaDataRequest.setEndpoint(instanceURL+'/api/now/attachment?sysparm_query=table_name%3D'+sourceTable+'%5EORtable_name%3DZZ_YY'+sourceTable+'%5Etable_sys_id%3D'+sourceSysId);
    getMetaDataRequest.setRequestHeader("Accept","application/json");
    getMetaDataRequest.setRequestHeader("Authorization", "Basic " + authorization);
    var getMetaDataResponse = getMetaDataRequest.execute();
    if(getMetaDataResponse.getStatusCode()!==200 && getMetaDataResponse.getStatusCode()!==201) {
        throw new Error("Failed to obtain source attachment meta data.\n Status: " + getMetaDataResponse.getStatusCode()+" Error: "+ getMetaDataResponse.getErrorMessage());
    }
    var jsonResponse=global.JSON.parse(getMetaDataResponse.getBody());
    for (var i=0;i<jsonResponse.result.length;i++) {
        var val=jsonResponse.result[i];
        //Attachments in the imageFieldsMap list (if defined) should be recorded as images
        var fileName=val.file_name;
        attachmentMetaData[val.sys_id]={};
        if(typeof imageFieldsMap!=="undefined" && typeof (imageFieldsMap[fileName])!=="undefined") {
            attachmentMetaData[val.sys_id].isImage=true;
        }
        attachmentMetaData[val.sys_id].file_name=fileName;
        attachmentMetaData[val.sys_id].size_bytes=val.size_bytes;
        attachmentMetaData[val.sys_id].content_type=val.content_type;
        attachmentMetaData[val.sys_id].sys_id=val.sys_id;
    }
    return attachmentMetaData;
}

//returns a list of source attachments and images associated with this record
function getAttachmentMetaData_Glide(response, sourceTable, sourceSysId,imageFieldsMap) {
    var grSourceAttachment=new GlideRecord("sys_attachment");
    grSourceAttachment.addEncodedQuery('table_name='+sourceTable+'^ORtable_name=ZZ_YY'+sourceTable+'^table_sys_id='+sourceSysId);
    grSourceAttachment.query();
    var attachmentMetaData={};
    while(grSourceAttachment.next()) {
		var fileName=grSourceAttachment.getValue("file_name");
        attachmentMetaData[grSourceAttachment.getValue("sys_id")]={};
        if(typeof imageFieldsMap!=="undefined" && typeof (imageFieldsMap[fileName])!=="undefined") {
            attachmentMetaData[grSourceAttachment.getValue("sys_id")].isImage=true;
        }
        attachmentMetaData[grSourceAttachment.getValue("sys_id")].file_name=fileName;
        attachmentMetaData[grSourceAttachment.getValue("sys_id")].size_bytes=grSourceAttachment.getValue("size_bytes");
        attachmentMetaData[grSourceAttachment.getValue("sys_id")].content_type=grSourceAttachment.getValue("content_type");
        attachmentMetaData[grSourceAttachment.getValue("sys_id")].sys_id=grSourceAttachment.getValue("sys_id");
    }
    return attachmentMetaData;
}

function setAttachment_REST(response, sourceURL,attachmentSysId,authorization,targetTableName,targetSysID,fileName,isImage) {
    try {
        var request = new sn_ws.RESTMessageV2();
        request.setHttpMethod('GET');
        request.setEndpoint(sourceURL+'/api/now/attachment/'+attachmentSysId+'/file');
        request.setRequestHeader("Accept","*/*");
        request.setRequestHeader("Authorization", "Basic " + authorization);
        request.saveResponseBodyAsAttachment(targetTableName,targetSysID,fileName);
        var getResponse = request.execute();
        var sysID=getResponse.getResponseAttachmentSysid();
		if(isImage)setImageTableName(sysID);
        if(getResponse.getStatusCode()===200 || getResponse.getStatusCode()===201 && typeof sysID!=="undefined") {
			
            return sysID;
        }
        else {
            throw new Error("Failed to migrate attachment "+fileName+" to target entry " + targetSysID +".\n Status: " + getResponse.getStatusCode()+" Error: "+ getResponse.getErrorMessage());
        }
    }
    catch(e) {
        throw e;
    }
}

function setAttachment_Glide(response, glideSysAttachment,grTarget,sourceAttachID,contentType,targetFileName,isImage) {
    try {
        var updatedAttachStrm=glideSysAttachment.getContentStream(sourceAttachID);  
        var updatedAttachment=glideSysAttachment.writeContentStream(grTarget,targetFileName,contentType,updatedAttachStrm);
        if(isImage)setImageTableName(updatedAttachment);
        return updatedAttachment;
    }
    catch(e) {
        throw e;
    }  
}

function arrayContains(a, obj) {
    for (var i = 0; i < a.length; i++) {
        if (a[i] === obj) {
            return true;
        }
    }
    return false;
}

function setImageTableName(attachID ){
    var att = new GlideRecord("sys_attachment");    
    att.addQuery("sys_id", attachID);
    att.orderByDesc("sys_created_on");
    att.query();
    if(att.next()){
        att.table_name = "ZZ_YY"+att.table_name;
        att.update();
    }
}

function getImageFieldListGlide(response,tableName) {
    var map={};
    var gr=new GlideRecord("sys_dictionary");
    gr.addEncodedQuery("name="+tableName);
    gr.addEncodedQuery("internal_type=user_image");
	gr.query();
	while(gr.next()) {
        map[gr.getValue("element")]=gr.getValue("element");
    }
    return map;
}
//gets a list of all the images on a given table using the REST API
function getImageFieldListREST(response,instanceURL,authorization, tableName) {
    var map={};
    var getMetaDataRequest = new sn_ws.RESTMessageV2();
    getMetaDataRequest.setHttpMethod('GET');
    getMetaDataRequest.setEndpoint(instanceURL+'/api/now/table/sys_dictionary?sysparm_query=name%3D'+tableName+'%5Einternal_type%3Duser_image&sysparm_fields=element');
    getMetaDataRequest.setRequestHeader("Accept","application/json");
    getMetaDataRequest.setRequestHeader("Authorization", "Basic " + authorization);
    var getMetaDataResponse = getMetaDataRequest.execute();
    if(getMetaDataResponse.getStatusCode()!==200 && getMetaDataResponse.getStatusCode()!==201) {
        throw new Error("Failed to obtain image field list.\n Status: " + getMetaDataResponse.getStatusCode()+" Error: "+ getMetaDataResponse.getErrorMessage());
    }
    var jsonResponse=global.JSON.parse(getMetaDataResponse.getBody());
    for (var i=0;i<jsonResponse.result.length;i++) {
        var val=jsonResponse.result[i];
        map[val.element]=val.element;
    }
    return map;
}

function buildImageFieldMap(sourceMappedImageFields,targetMappedImageFields) {
    var sourceImageFields = sourceMappedImageFields.split(",");
    var targetImageFields = targetMappedImageFields.split(",");
    var imageFieldMap={};
    if(sourceImageFields.length!==targetImageFields.length) {
        throw new Error("Number of target image fields ["+ targetImageFields.length +"] does not match number of source image fields [" + sourceImageFields.length+"]")
    }
    for(var j=0;j<sourceImageFields.length;j++) {
        imageFieldMap[sourceImageFields[j]]=targetImageFields[j];
    }
    return imageFieldMap;		
}

function getAttachmentWithName(attachmentMap,name) {
	for(var key in attachmentMap) {
        var attachRecord=attachmentMap[key];
		if(attachRecord.file_name===name) {
			return attachRecord;
		}
	}
	return undefined;
}